/*
 * This source file contains the code for proxying calls in the master thread to calls in the workers
 * by `.postMessage()`-ing.
 *
 * Keep in mind that this code can make or break the program's performance! Need to optimize more…
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import isSomeObservable from "is-observable";
import { multicast, Observable } from "observable-fns";
import { CommonMessageType } from "../types/messages";
import { lookupLocalCallback } from "./callbacks";
import { ObservablePromise } from "./observable-promise";
import { isTransferDescriptor } from "./transferable";
let nextCallID = 1;
const activeSubscriptions = new Map();
const dedupe = (array) => Array.from(new Set(array));
const isCallCancelMessage = (data) => data && data.type === CommonMessageType.cancel;
const isCallErrorMessage = (data) => data && data.type === CommonMessageType.error;
const isCallResultMessage = (data) => data && data.type === CommonMessageType.result;
const isCallRunningMessage = (data) => data && data.type === CommonMessageType.running;
const isInvocationMessage = (data) => data && data.type === CommonMessageType.invoke;
function isZenObservable(thing) {
    return thing && typeof thing === "object" && typeof thing.subscribe === "function";
}
/**
 * There are issues with `is-observable` not recognizing zen-observable's instances.
 * We are using `observable-fns`, but it's based on zen-observable, too.
 */
function isObservable(thing) {
    return isSomeObservable(thing) || isZenObservable(thing);
}
function deconstructTransfer(thing) {
    return isTransferDescriptor(thing)
        ? { payload: thing.send, transferables: thing.transferables }
        : { payload: thing, transferables: undefined };
}
function postCallError(relay, uid, rawError) {
    const { payload: error, transferables } = deconstructTransfer(rawError);
    const errorMessage = {
        type: CommonMessageType.error,
        uid,
        error
    };
    relay.postMessage(errorMessage, transferables);
}
function postCallResult(relay, uid, completed, resultValue) {
    const { payload, transferables } = deconstructTransfer(resultValue);
    const resultMessage = {
        type: CommonMessageType.result,
        uid,
        complete: completed ? true : undefined,
        payload
    };
    relay.postMessage(resultMessage, transferables);
}
function postCallRunning(relay, uid, resultType) {
    const startMessage = {
        type: CommonMessageType.running,
        uid,
        resultType
    };
    relay.postMessage(startMessage);
}
function createObservableForJob(relay, serializer, callID, debug) {
    return new Observable(observer => {
        let asyncType;
        const messageHandler = ((event) => {
            const message = event.data;
            if (!message || message.uid !== callID)
                return;
            debug(`Received message for running call ${callID}:`, message);
            if (isCallRunningMessage(message)) {
                asyncType = message.resultType;
            }
            else if (isCallResultMessage(message)) {
                if (asyncType === "promise") {
                    if (typeof message.payload !== "undefined") {
                        observer.next(serializer.deserialize(message.payload, relay));
                    }
                    observer.complete();
                    relay.removeEventListener("message", messageHandler);
                }
                else {
                    if (message.payload) {
                        observer.next(serializer.deserialize(message.payload, relay));
                    }
                    if (message.complete) {
                        observer.complete();
                        relay.removeEventListener("message", messageHandler);
                    }
                }
            }
            else if (isCallErrorMessage(message)) {
                const error = serializer.deserialize(message.error, relay);
                if (asyncType === "promise" || !asyncType) {
                    observer.error(error);
                }
                else {
                    observer.error(error);
                }
                relay.removeEventListener("message", messageHandler);
            }
        });
        relay.addEventListener("message", messageHandler);
        return () => {
            if (asyncType === "observable" || !asyncType) {
                const cancelMessage = {
                    type: CommonMessageType.cancel,
                    uid: callID
                };
                relay.postMessage(cancelMessage);
            }
            relay.removeEventListener("message", messageHandler);
        };
    });
}
function prepareArguments(serializer, rawArgs) {
    if (rawArgs.length === 0) {
        // Exit early if possible
        return {
            args: [],
            transferables: []
        };
    }
    const args = [];
    const transferables = [];
    for (const arg of rawArgs) {
        if (isTransferDescriptor(arg)) {
            args.push(serializer.serialize(arg.send));
            transferables.push(...arg.transferables);
        }
        else {
            args.push(serializer.serialize(arg));
        }
    }
    return {
        args,
        transferables: transferables.length === 0 ? transferables : dedupe(transferables)
    };
}
export function createProxyFunction(relay, serializer, fid, debug) {
    return ((...rawArgs) => {
        const uid = nextCallID++;
        const { args, transferables } = prepareArguments(serializer, rawArgs);
        const runMessage = {
            type: CommonMessageType.invoke,
            fid,
            uid,
            args
        };
        debug("Sending command to run function to worker:", runMessage);
        try {
            relay.postMessage(runMessage, transferables);
        }
        catch (error) {
            return ObservablePromise.from(Promise.reject(error));
        }
        return ObservablePromise.from(multicast(createObservableForJob(relay, serializer, uid, debug)));
    });
}
export function createProxyModule(relay, serializer, methods, debug) {
    const proxy = {};
    for (const methodName of Object.keys(methods)) {
        proxy[methodName] = createProxyFunction(relay, serializer, methods[methodName], debug);
    }
    return proxy;
}
function invokeExposedLocalFunction(relay, serializer, callback, message) {
    return __awaiter(this, void 0, void 0, function* () {
        let syncResult;
        const uid = message.uid;
        try {
            const args = message.args.map(arg => serializer.deserialize(arg, relay));
            syncResult = callback(...args);
        }
        catch (error) {
            postCallError(relay, uid, serializer.serialize(error));
        }
        const resultType = isObservable(syncResult) ? "observable" : "promise";
        postCallRunning(relay, uid, resultType);
        if (isObservable(syncResult)) {
            const subscription = syncResult.subscribe(value => postCallResult(relay, uid, false, serializer.serialize(value)), error => postCallError(relay, uid, serializer.serialize(error)), () => postCallResult(relay, uid, true));
            activeSubscriptions.set(uid, subscription);
        }
        else {
            try {
                const result = yield syncResult;
                postCallResult(relay, uid, true, serializer.serialize(result));
            }
            catch (error) {
                postCallError(relay, uid, serializer.serialize(error));
            }
        }
    });
}
function handleRemoteInvocation(relay, serializer, message, debug) {
    const callback = lookupLocalCallback(message.fid);
    if (!callback) {
        debug(`Call to exposed local function failed: Function not found: UID ${message.uid}`);
        return postCallError(relay, message.uid, serializer.serialize(Error(`Function not found: UID ${message.uid}`)));
    }
    debug(`Received invocation of local exposed function ${message.fid}, call UID ${message.uid} with arguments:`, message.args);
    return invokeExposedLocalFunction(relay, serializer, callback, message);
}
export function handleFunctionInvocations(relay, serializer, debug) {
    relay.addEventListener("message", (event) => {
        debug(`Received message:`, event.data);
        if (isInvocationMessage(event.data)) {
            handleRemoteInvocation(relay, serializer, event.data, debug);
        }
    });
}
export function handleCallCancellations(relay, debug) {
    relay.addEventListener("message", event => {
        const messageData = event.data;
        if (isCallCancelMessage(messageData)) {
            const jobUID = messageData.uid;
            const subscription = activeSubscriptions.get(jobUID);
            if (subscription) {
                subscription.unsubscribe();
                activeSubscriptions.delete(jobUID);
            }
        }
    });
}
