"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.expose = exports.isWorkerRuntime = void 0;
const debug_1 = __importDefault(require("debug"));
const call_proxy_1 = require("../common/call-proxy");
const callbacks_1 = require("../common/callbacks");
const serialization_1 = require("../common/serialization");
const messages_1 = require("../types/messages");
const implementation_1 = __importDefault(require("./implementation"));
var callbacks_2 = require("../common/callbacks");
Object.defineProperty(exports, "Callback", { enumerable: true, get: function () { return callbacks_2.Callback; } });
var serialization_2 = require("../common/serialization");
Object.defineProperty(exports, "registerSerializer", { enumerable: true, get: function () { return serialization_2.registerSerializer; } });
var transferable_1 = require("../common/transferable");
Object.defineProperty(exports, "Transfer", { enumerable: true, get: function () { return transferable_1.Transfer; } });
/** Returns `true` if this code is currently running in a worker. */
exports.isWorkerRuntime = implementation_1.default.isWorkerRuntime;
let exposeCalled = false;
const debugIncomingMessages = debug_1.default("threads:worker:messages");
function postFunctionInitMessage() {
    const initMessage = {
        type: messages_1.WorkerMessageType.init,
        exposed: {
            type: "function"
        }
    };
    implementation_1.default.postMessage(initMessage);
}
function postModuleInitMessage(methods) {
    const initMessage = {
        type: messages_1.WorkerMessageType.init,
        exposed: {
            type: "module",
            methods
        }
    };
    implementation_1.default.postMessage(initMessage);
}
function postUncaughtErrorMessage(error) {
    try {
        const errorMessage = {
            type: messages_1.WorkerMessageType.uncaughtError,
            error: serialization_1.serialize(error)
        };
        implementation_1.default.postMessage(errorMessage);
    }
    catch (subError) {
        // tslint:disable-next-line no-console
        console.error("Not reporting uncaught error back to master thread as it " +
            "occured while reporting an uncaught error already." +
            "\nLatest error:", subError, "\nOriginal error:", error);
    }
}
/**
 * Expose a function or a module (an object whose values are functions)
 * to the main thread. Must be called exactly once in every worker thread
 * to signal its API to the main thread.
 *
 * @param exposed Function or object whose values are functions
 */
function expose(exposed) {
    if (!implementation_1.default.isWorkerRuntime()) {
        throw Error("expose() called in the master thread.");
    }
    if (exposeCalled) {
        throw Error("expose() called more than once. This is not possible. Pass an object to expose() if you want to expose multiple functions.");
    }
    exposeCalled = true;
    if (typeof exposed === "function") {
        callbacks_1.SingleExposedCallback(exposed);
        call_proxy_1.handleFunctionInvocations(implementation_1.default, serialization_1.getRegisteredSerializer(), debugIncomingMessages);
        postFunctionInitMessage();
    }
    else if (typeof exposed === "object" && exposed) {
        const methods = Object.keys(exposed).reduce((reduced, methodName) => {
            const callback = callbacks_1.Callback(exposed[methodName]);
            return Object.assign(Object.assign({}, reduced), { [methodName]: callback.id });
        }, {});
        call_proxy_1.handleFunctionInvocations(implementation_1.default, serialization_1.getRegisteredSerializer(), debugIncomingMessages);
        postModuleInitMessage(methods);
    }
    else {
        throw Error(`Invalid argument passed to expose(). Expected a function or an object, got: ${exposed}`);
    }
    call_proxy_1.handleCallCancellations(implementation_1.default, debugIncomingMessages);
}
exports.expose = expose;
if (typeof self !== "undefined" && typeof self.addEventListener === "function" && implementation_1.default.isWorkerRuntime()) {
    self.addEventListener("error", event => {
        // Post with some delay, so the master had some time to subscribe to messages
        setTimeout(() => postUncaughtErrorMessage(event.error || event), 250);
    });
    self.addEventListener("unhandledrejection", event => {
        const error = event.reason;
        if (error && typeof error.message === "string") {
            // Post with some delay, so the master had some time to subscribe to messages
            setTimeout(() => postUncaughtErrorMessage(error), 250);
        }
    });
}
if (typeof process !== "undefined" && typeof process.on === "function" && implementation_1.default.isWorkerRuntime()) {
    process.on("uncaughtException", (error) => {
        // Post with some delay, so the master had some time to subscribe to messages
        setTimeout(() => postUncaughtErrorMessage(error), 250);
    });
    process.on("unhandledRejection", (error) => {
        if (error && typeof error.message === "string") {
            // Post with some delay, so the master had some time to subscribe to messages
            setTimeout(() => postUncaughtErrorMessage(error), 250);
        }
    });
}
